/*******************************************

 SYSFUNCS.CPP - The KINT Interpreter
	Copyright 2002 Everett Kaser
	All rights reserved.

	Do not copy or redistribute without the
	author's explicit written permission.
 *******************************************/
#include <windows.h>
#include <stdio.h>

#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <mmsystem.h>

#include "main.h"
#include "pmach.h"
#include "sysfuncs.h"

//************************************************************
//* PORTING NOTES
//* -------------
//* 1) See PORTING NOTES at start of PMACH.CPP
//* 2) Kode (at this point) assumes that the screen size in
//*		pixels ranges from 640x480 to 1600x1280.  Larger (or
//*		smaller sizes may be supported, but that gives you a
//*		range that you may have to "scale to" if your actual
//*		display size is greatly different (like 10-20 years
//*		from now :-).
//************************************************************

extern HANDLE	ghInstance;
extern HWND		hWndMain;
extern char		PgmPath[512], TmpPath[512];
extern BYTE		Kregs[256]; // 4*32 + 2*48 + 1*32 = 256 (32 DWORDS, 48 WORDS, 32 BYTES)
extern WORD		MusicPlaying;
extern BYTE		gateEvent;
extern int		ScreenWidth, ScreenHeight;
extern int		CurCur, mainShowCmd;
extern WORD		ScreenWidthIndex;
extern UINT		TimerID;

//*************************************************************

UINT MCIwDeviceID;
MCI_OPEN_PARMS mciOpenParms;
MCI_PLAY_PARMS mciPlayParms;
MCI_STATUS_PARMS mciStatusParms;

char	*pMSCmem, *pNxtSrc;
DWORD	fsize, nread;

int	bitspixel, planes, numcolors, sizepalette, colormode;
HPALETTE	kPal;
void DestroyPalette();

DOCINFO PRTdocinfo = {sizeof(DOCINFO), "EKS puzzle", NULL, NULL, 0};
HDC		PRThdc=NULL;

PDIBS		pFirstBM=NULL, kBM=NULL;
HDC			ddc, sdc;
HBITMAP		oldBMs, oldBMd;
int			defFont=-1;

#define		MAXFONTS	30
FONTARRAY	Fonts[MAXFONTS];

OLDMODES	Moldie[MAXFONTS];
int			MoldCnt=0;

//***************************************
// Gets a Device Context for printing
HDC GetPrinterDC(void) {

	HDC				hdc;
	static PAGESETUPDLG	psd={
		sizeof(PAGESETUPDLG), NULL, NULL, NULL,
		PSD_DISABLEMARGINS | PSD_DISABLEPAGEPAINTING | PSD_DISABLEPAPER | PSD_INTHOUSANDTHSOFINCHES,
		};
	char	*pNames;

	psd.hwndOwner = hWndMain;
	psd.hInstance = ghInstance;
	psd.lCustData = NULL;
	psd.lpfnPageSetupHook = NULL;
	psd.lpfnPagePaintHook = NULL;
	psd.lpPageSetupTemplateName = NULL;
	psd.hPageSetupTemplate = NULL;
	if( PageSetupDlg(&psd) ) {
		pNames = (char *)GlobalLock(psd.hDevNames);
		hdc = CreateDC(	pNames + ((DEVNAMES *)pNames)->wDriverOffset,
						pNames + ((DEVNAMES *)pNames)->wDeviceOffset, NULL, NULL);
		GlobalUnlock(psd.hDevNames);
	} else hdc = NULL;
	return hdc;
}

//*****************************************************************************
/* Plays a given MIDI file using MCI_OPEN, MCI_PLAY. Returns as soon as
 * playback begins. The window procedure function for the given window
 * will be notified when playback is complete. Returns 0L on success;
 * otherwise, it returns an MCI error code.
 */
MCIERROR playMIDIFile(HWND hWndNotify, const char *lpszMIDIFileName) {

    MCIERROR dwReturn;

	if( MusicPlaying ) stopMIDIFile();
// Open the device by specifying the device name and device element.
// MCI will attempt to choose the MIDI Mapper as the output port.
	mciOpenParms.dwCallback = 0;
	mciOpenParms.wDeviceID = 0;
    mciOpenParms.lpstrDeviceType = NULL;
    mciOpenParms.lpstrElementName = lpszMIDIFileName;
	mciOpenParms.lpstrAlias = NULL;
    if( dwReturn=mciSendCommand(NULL, MCI_OPEN, MCI_WAIT | MCI_OPEN_ELEMENT, (DWORD)(LPVOID) &mciOpenParms)) {
// Failed to open device; don't close it, just return error.
        return (dwReturn);
    }
// Device opened successfully, get the device ID.
    MCIwDeviceID = mciOpenParms.wDeviceID;
// Begin playback. The window procedure function for the parent window
// will be notified with an MM_MCINOTIFY message when playback is
// complete. At that time, the window procedure closes the device.
    mciPlayParms.dwCallback = (DWORD)(hWndNotify);
    if( dwReturn=mciSendCommand(MCIwDeviceID, MCI_PLAY, MCI_NOTIFY, (DWORD)(LPVOID) &mciPlayParms)) {
        mciSendCommand(MCIwDeviceID, MCI_CLOSE, 0, NULL);
        return( dwReturn );
    }
	MusicPlaying = 1;
    return( 0L );
}


//*********************************************************
void stopMIDIFile() {

	mciSendCommand(MCIwDeviceID, MCI_STOP, MCI_WAIT, NULL);
	mciSendCommand(MCIwDeviceID, MCI_CLOSE, MCI_WAIT, NULL);
	MusicPlaying = 0;
}

//*********************************************************
void BuildPath(BYTE *pMem, BYTE folder) {

	strcpy(TmpPath, PgmPath);
	switch( folder ) {
	 case 0:	// GAME (home) folder
		strcat(TmpPath, (char *)pMem);
		break;
	 case 1:
		strcat(TmpPath, "UI\\");
		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".bmp");
		break;
	 case 2:
		strcat(TmpPath, "IMAGES\\");
		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".bmp");
		break;
	 case 3:
		strcat(TmpPath, "BACKGNDS\\");
		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".bmp");
		break;
	 case 4:
	// note: in this implementation, all music files are in a "common" folder up one level from the 'home' folder
	// so that other games from Everett Kaser Software can share the same MIDI files.
		strcat(TmpPath, "..\\MUSIC\\");
		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".mid");
		break;
	 case 5:
	// note: in this implementation, we store the WAV files in the 'home' folder
		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".wav");
		break;
	 default:
		break;
	}
}

//************************************************************
//* In general, DWORD arguments start with R15 and progress downwards
//*		WORD arguments start with R39 and progress downwards
//*		BYTE arguments start with R95 and progress downwards
//*		return values are generally in R15, R39, and/or R95
//*		of course, there's always exceptions to every rule...
void SysFunc(BYTE *Kregs, WORD opcode) {

	HPEN		hPen, holdPen;
	HBRUSH		hBrush, holdBrush;
	PDIBS		pBM, pBM2;
	RECT		rect;
	BYTE		*pMem, *pMem2;
	_finddata_t	fdat;
	time_t		curtime;
	struct	tm	*ourtime;
	POINT		pts[5];
	int			i, oldstretchs, oldstretchd, iread, *pI;
	DWORD		dw1, dw2;
	WORD		w, h;
	WINDOWPLACEMENT	WndPlace;
	char		*pChar;
	SIZE		txtsize;
	TEXTMETRIC	tm;
	HFONT		hfo;
	UINT		showCmd;
	HANDLE		hFile;

	switch( opcode ) {
//**** SYS_FILESIZE
// NOTE: doesn't worry about upper DWORD of GetFileSize, because no files
//	used by this system are ever that large.
// inputs:
//	R15 = offset of filename in memblk [R95]
//	R95 = memblk of filename
//	R94 = folder, as in:
//		0 = folder the game is installed in (home)
//		1 = UI (folder containing UI image files)
//		2 = IMAGES (folder containing game image set files)
//		3 = BACKGNDS ( folder containing 'decorative' bitmap files)
//		4 = MUSIC (folder containing music files)
//		5 = SOUNDS (folder containing sound effect files)
// outputs:
//	R13 = size of file (-1 if error)
	 case SYS_FILESIZE:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		BuildPath(pMem, GETREGB(REG_94));
		hFile = CreateFile(TmpPath, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if( hFile==INVALID_HANDLE_VALUE ) SETREGD(REG_13, 0xFFFFFFFF);
		else {
			SETREGD(REG_13, GetFileSize(hFile, NULL));
			CloseHandle(hFile);
		}
		return;
//**** SYS_OPEN:
// inputs:
//	R15 = offset of filename in memory block specified by R95
//	R95 = memory block # of filename (0 if R0 based, 1 if R1 based, 2 if R2 based)
//	R94 = folder, as in:
//		0 = folder the game is installed in (home)
//		1 = UI (folder containing UI image files)
//		2 = IMAGES (folder containing game image set files)
//		3 = BACKGNDS ( folder containing 'decorative' bitmap files)
//		4 = MUSIC (folder containing music files)
//		5 = SOUNDS (folder containing sound effect files)
//
//	In the Windows implementation, UI, IMAGES, and BACKGNDS files
//	are all .BMP files, and get that file extendsion added to them
//	automatically.  MUSIC files are assumed to be MIDI files and get .MID.
//	SOUNDS are assumed to be .WAV files, and get .WAV appended.
//	When porting this to another platform (or even under Windows),
//	there's absolutely NO reason why you can't use other graphics file
//	formats (so long as they're 'lossless', so as to preserve the exact
//	colors of the pixels, so that transparency works right), or other
//	music (like MP3) or sound file formats.  As long as you modify
//	the KINT code that opens and loads/plays the files, you can use
//	whatever file formats you want.  The KASM code doesn't care.
//
//	R39 = open flags.  Bits are:
//		 _O_RDONLY	0x0000	/* open for reading only */
//		 _O_RDONLY	0x0000	/* open for reading only */
//		 _O_WRONLY	0x0001	/* open for writing only */
//		 _O_RDWR 	0x0002	/* open for reading and writing */
//		 _O_APPEND	0x0008	/* writes done at eof */
//		 _O_CREAT	0x0100	/* create and open file */
//		 _O_TRUNC	0x0200	/* open and truncate */
//		 _O_TEXT 	0x4000	/* file mode is text (translated) */
//		 _O_BINARY	0x8000	/* file mode is binary (untranslated) */
// outputs:
//	R15 = file handle if successful, -1 if failed
	 case	SYS_FOPEN:
		// WARN: PTR4
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		BuildPath(pMem, GETREGB(REG_94));
		SETREGD(REG_15, (DWORD)_open(TmpPath, (int)(DWORD)GETREGW(REG_39), _S_IREAD | _S_IWRITE));
		return;
//**** SYS_EOF
// Inputs:
//	R15 = file handle
// Outputs:
//	R12 = 1 if at EOF, 0 if not at EOF, -1 if error
	 case	SYS_FEOF:
		SETREGD(REG_12, (DWORD)_eof((int)GETREGD(REG_15)));
		return;
//**** SYS_READ
// Inputs:
//	R15 = file handle
//	R14 = buffer offset
//	R13 = max number of bytes to read
//	R95 = memory block # (0 if R0 based, 1 if R1 based, 2 if R2 based)
// Outputs:
//	R15 = number of bytes read.  (-1 if error)
	 case	SYS_FREAD:
		// WARN: PTR4
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_14);
		i = (int)GETREGD(REG_15);	// file handle
		dw1 = GETREGD(REG_13);
		dw2 = 0;
		while( dw1 ) {
			if( dw1<16384 ) w = (WORD)dw1;
			else w = 16384;
			iread = _read(i, (void *)pMem, (int)w);
			if( iread==-1 ) {
				dw2 = -1;
				break;
			}
			if( iread==0 ) break;
			dw2 += (DWORD)iread;
			pMem += iread;
			dw1 -= (DWORD)iread;
		}
		SETREGD(REG_12, dw2);
		return;
//**** SYS_WRITE
// Inputs:
//	R15 = file handle
//	R14 = buffer offset in memory block specified by R95
//	R13 = number of bytes to write
//	R95 = memory block # (0 if R0 based, 1 if R1 based, 2 if R2 based)
// Outputs:
//	R12 = number of bytes written.  (-1 if error)
	 case	SYS_FWRITE:
		// WARN: PTR4
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_14);
		i = (int)GETREGD(REG_15);	// file handle
		dw1 = GETREGD(REG_13);		// number of bytes to write
		dw2 = 0;					// number written
		while( dw1 ) {
			if( dw1<16384 ) w = (WORD)dw1;
			else w = 16384;
			iread = _write(i, (void *)pMem, (int)(DWORD)w);
			if( iread==-1 ) {
				dw2 = -1;
				break;
			}
			dw2 += (DWORD)iread;
			pMem += iread;
			dw1 -= (DWORD)iread;
		}
		SETREGD(REG_12, dw2);
		return;
//**** SYS_SEEK
// Inputs:
//	R15 = file handle
//	R14 = offset to move the file pointer to
//	R95 = origin:
//		SEEK_SET    0
//		SEEK_CUR    1
//		SEEK_END    2
// Outputs:
//	R12 = offset in bytes from the beginning of the file of the new position (-1 if error)
	 case	SYS_FSEEK:
		SETREGD(REG_12, (DWORD)_lseek((int)GETREGD(REG_15), (long)GETREGD(REG_14), (int)GETREGB(REG_95)));
		return;
//**** SYS_CLOSE
// Inputs:
//	R15 = file handle to close
// Outputs:
//	R12 = 0 if successful, -1 if failed
	 case	SYS_FCLOSE:
		_close((int)GETREGD(REG_15));
		return;
//**** SYS_FINDFIRST
// Inputs:
//	R15 = offset in memblock[R95] of filespec to search for
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblock[R94] of buffer for found filename (must be minimum of 260 bytes)
//	R94 = 0 for R0, 1 for R1, 2 for R2
//	R93 = folder, as in:
//		0 = folder the game is installed in (home)
//		1 = UI (folder containing UI image files)
//		2 = IMAGES (folder containing game image set files)
//		3 = BACKGNDS ( folder containing 'decorative' bitmap files)
//		4 = MUSIC (folder containing music files)
//		5 = SOUNDS (folder containing sound effect files)
// Outputs:
//	R12 = "findfirst/next" handle of first file found, or -1 if none found or error
//	  and found file name in buffer [R94]:R14
//  R95 = file attributes:
//		 attributes of file found: 0x00 = _A_NORMAL, 0x10 = _A_SUBDIR
//		NOTE: the returned file name must be JUST the filename, not
//			including any part of the full path or the file extension.
	 case	SYS_FFINDFIRST:
	 // WARN: PTR4
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		BuildPath(pMem, GETREGB(REG_93));
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		SETREGD(REG_15, (DWORD)_findfirst(TmpPath, &fdat));
findnextcom:
		SETREGB(REG_95, (BYTE)fdat.attrib);
		if( GETREGD(REG_15) != 0xFFFFFFFF ) {
		// find JUST the filename portion
			for(pChar=fdat.name+strlen(fdat.name); pChar>fdat.name && *(pChar-1)!='\\'; --pChar) ;
			strcpy((char *)pMem2, pChar);
			if( GETREGB(REG_93) ) {
				for(pChar=(char *)pMem2+strlen((char *)pMem2); pChar>(char *)pMem2 && *pChar!='.'; --pChar) ;
				if( *pChar=='.' ) *pChar = '\0';	// remove file extension from all but R93==0 case
			}
		}
		return;
//**** SYS_FINDNEXT
// Inputs:
//	R15 = "find first/next" handle returned by previous FINDFIRST
//	R14 = offset in memblock[R94] of buffer for found filename (must be minimum of 260 bytes)
//	R94 = 0 for R0, 1 for R1, 2 for R2
// Outputs:
//	R15 = "findfirst/next" handle of file found, or -1 if none found or error
//	  and found file name in buffer [R94]:R14
//  R95 = file attributes:
//		 attributes of file found: 0x00 = _A_NORMAL, 0x10 = _A_SUBDIR
	 case	SYS_FFINDNEXT:
	 // WARN: PTR4
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		SETREGD(REG_15, (DWORD)_findnext((long)GETREGD(REG_15), &fdat));
		goto findnextcom;
//**** SYS_FINDCLOSE
// Inputs:
//	R15 = "find first/next" handle returned by previous FINDFIRST
	 case	SYS_FFINDCLOSE:
		_findclose((long)GETREGD(REG_15));
		return;
//**** SYS_DELETE
// Inputs:
//	R15 = address of filename
//	R95 = memblk of filename
	 case	SYS_DELETE:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		strcpy(TmpPath, PgmPath);
		strcat(TmpPath, (char *)pMem);
		_unlink(TmpPath);
		return;
//**** SYS_MAKESHARECOPY
// Inputs:
//	R95 = command
//	R15 = address of buffer in memblk R0 for next filename
//	R13 = address of buffer in memblk R0 containing destination path (if R95 != 0)
// Outputs:
//	R14 = -1 if error
//		else total size of files to be copied if cmd==0 (and filename of FIRST file in R15 buffer)
//		else 0 (and filename of NEXT file in R15 buffer, or NUL if last file)
//
// NOTE: KASM code always calls this FIRST with command=0 (read file names, return first file name in buffer R15 and total size in R14)
//  followed by a sequential series of calls with R95 != 0 to copy successive files.
//  each successive call returns the NEXT filename to be copied in buffer R15.
//	When no more files to be copied, buffer R15 starts with a NUL.

// NOTE: this function is FAR more complex than ordinarily need be, but I didn't want
// the KASM code to know anything about the files being copied for the shareware version
// (as their names and/or formats will undoubtedly be different on different OS's), and
// I didn't want the KINT interpreter (this code) to know anything about the specific
// list of files being copied either, as I want the KINT code to remain generic across
// all of the games.  So, the files to be copied are in an ASCII file call "swlist",
// as a comma-separated list of src/dst pairs.  For example:
//	eks0.ebk,setup.exe,eksctlss.ebk,setup.ctl,ekssubk.ebk,setupbk.ebk,shs.epk,shs.epk
// all on one line, no CRs, no LFs, source filename first, comma, destination filename
// second, comma, next pair....etc.
//
	 case	SYS_MAKESHARECOPY:
		pMem = (BYTE *)GETREGD(REG_0) + GETREGD(REG_15);
		if( GETREGB(REG_95)==0 ) {	// get total sizes and return first filename in R15 buffer
			pMSCmem = NULL;
			strcpy(TmpPath, PgmPath);
			strcat(TmpPath, "swlist");

			hFile = CreateFile(TmpPath, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if( hFile==INVALID_HANDLE_VALUE ) {
MSCerr:
				if( pMSCmem!=NULL ) free(pMSCmem);
				if( hFile!=INVALID_HANDLE_VALUE ) CloseHandle(hFile);
				SETREGD(REG_14, 0xFFFFFFFF);
			} else {
				fsize = GetFileSize(hFile, NULL);
				pMSCmem = (char *)malloc(fsize+1);
				if( pMSCmem==NULL ) goto MSCerr;
				ReadFile(hFile, pMSCmem, fsize, &nread, NULL);
				CloseHandle(hFile);
				for(pMem2=(BYTE *)pMSCmem; (char *)pMem2<(pMSCmem+nread); pMem2++) if( *pMem2==',' ) *pMem2 = 0;
				*pMem2 = 0;	// make sure last entry is nul-terminated
				dw1 = dw2 = 0;
				for(pMem2=(BYTE *)pMSCmem; (char *)pMem2<(pMSCmem+nread); ) {
					strcpy(TmpPath, PgmPath);
					strcat(TmpPath, (const char *)pMem2);	// next src filename

					hFile = CreateFile(TmpPath, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
					if( hFile==INVALID_HANDLE_VALUE ) goto MSCerr;
					dw1 += (GetFileSize(hFile, NULL)+8191) & 0xFFFFE000;	// add it's size to the running total (rounded up to nearest 8Kbytes for estimated file system slop)
					CloseHandle(hFile);

					while( *pMem2!=0 && (char *)pMem2<(pMSCmem+nread) ) ++pMem2;	// skip the src name
					++pMem2;	// skip the nul
					if( !dw2 ) strcpy((char *)pMem, (const char *)pMem2);	// if first one, copy dst name to R15 buffer
					++dw2;	// no longer on first one
					while( *pMem2!=0 && (char *)pMem2<(pMSCmem+nread) ) ++pMem2;	// skip the dst name
					++pMem2;	// skip the nul
				}
				pNxtSrc = pMSCmem;		// set us ready to copy first file
				SETREGD(REG_14, dw1);	// return total size of files
			}
		} else { // copying file
			char	target[256];

			BuildPath((BYTE *)pNxtSrc, 0);	// all shareware files to copy must be located in DIR_GAME

			strcpy(target, (char *)(GETREGD(REG_0)) + GETREGD(REG_13));
// make sure target path is properly terminated for this OS for concatenating the filename
// onto the end of it.  For MS Windows, that means making sure it ends in a '\'
			i = strlen(target);
			if( !i || target[i-1]!='\\' ) strcat(target, "\\");

			while( *pNxtSrc!=0 && pNxtSrc<(pMSCmem+nread) ) ++pNxtSrc;	// skip src filename
			++pNxtSrc;	// skip the nul
			strcat(target, pNxtSrc);	// finish building target filename
			while( *pNxtSrc!=0 && pNxtSrc<(pMSCmem+nread) ) ++pNxtSrc;	// skip dst filename
			if( pNxtSrc<(pMSCmem+nread) ) ++pNxtSrc;	// skip the nul if not the last one

			pMem2 = (BYTE *)pNxtSrc;
			while( *pMem2!=0 && (char *)pMem2<(pMSCmem+nread) ) ++pMem2;	// skip the dst name
			if( (char *)pMem2<(pMSCmem+nread) ) ++pMem2;	// skip the nul if this isn't the last one
			strcpy((char *)pMem, (const char *)pMem2);	// copy the next filename into R15 buffer (or nul if no more)

			if( CopyFile(TmpPath, target, FALSE) ) {	// if copy successful
				SETREGD(REG_14, 0);
				if( *pMem==0 ) free(pMSCmem);	// free memory if this was the last file to copy
			} else {
				SETREGD(REG_14, -1);
				free(pMSCmem);
			}
		}
		return;
//**** SYS_PRINTSTART
// Inputs:
// Outputs:
//	R15 = 0 if fail, otherwise non-0
//	R14 = page size in X direction (total pixels in printable area)
//	R13 = page size in Y direction (total pixels in printable area)
//	R12 = printer resolution in X direction (pixels per inch)
//	R11 = printer resolution in Y direction (pixels per inch)
	 case	SYS_PRINTSTART:
		if( PRThdc==NULL ) PRThdc = GetPrinterDC();

		SETREGD(REG_14, (DWORD)(long)GetDeviceCaps(PRThdc, HORZRES));
		SETREGD(REG_13, (DWORD)(long)GetDeviceCaps(PRThdc, VERTRES));
		SETREGD(REG_12, (DWORD)(long)GetDeviceCaps(PRThdc, LOGPIXELSX));
		SETREGD(REG_11, (DWORD)(long)GetDeviceCaps(PRThdc, LOGPIXELSY));

		if( StartDoc(PRThdc, &PRTdocinfo) <= 0 ) {
			DeleteDC(PRThdc);
			PRThdc = NULL;
		}
		
		SETREGD(REG_15, (DWORD)PRThdc);	// used by text and graphics output Kodes
		return;
//**** SYS_PRINTBEGPAGE
// Output:
//	R95 = 1 if successful, 0 if fail
	 case	SYS_PRINTBEGPAGE:
		if( StartPage(PRThdc) > 0 ) SETREGB(REG_95, 1);
		else SETREGB(REG_95, 0);
		return;
//**** SYS_PRINTENDPAGE
// Output:
//	R95 = 1 if successful, 0 if fail
	 case	SYS_PRINTENDPAGE:
		if( EndPage(PRThdc) > 0 ) SETREGB(REG_95, 1);
		else SETREGB(REG_95, 0);
		return;
//**** SYS_PRINTSTOP
	 case	SYS_PRINTSTOP:
		EndDoc(PRThdc);
		DeleteDC(PRThdc);
		PRThdc = NULL;
		return;
//**** SYS_PRINTTEXT
// inputs:
//	R14 = string buffer address in memblk[R95]
//	R13 = textcolor
//	R95 = memblk for string buffer
//	R12 = x-coordinate of point
//	R11 = y-coordinate of point
//	R37 = font handle

	 case	SYS_PRINTTEXT:
		SetBkMode(PRThdc, TRANSPARENT);
		SetTextColor(PRThdc, (COLORREF)GETREGD(REG_13));

		w = GETREGW(REG_37);
		if( w<MAXFONTS && Fonts[w].used ) SelectObject(PRThdc, Fonts[w].hFont);
		SetTextAlign(PRThdc, TA_LEFT | TA_TOP);

		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);

		SelectPalette(PRThdc, kPal, 0);
		RealizePalette(PRThdc);
		TextOut(PRThdc, (int)((long)GETREGD(REG_12)), (int)((long)GETREGD(REG_11)), (char *)pMem, strlen((char *)pMem));
		return;
//**** SYS_PRINTBM
// inputs:
//	R14 = handle of source bitmap
//	R19 = x-coord of destination
//	R18 = y-coord of destination
//	R17 = x-coord of source
//	R16 = y-coord of source
//	R13 = width of source
//	R12 = height of source
//	R11 = width of destination
//	R10 = height of destination
	 case	SYS_PRINTBM:
		pBM2 = (PDIBS)GETREGD(REG_14);
		if( pBM2==NULL ) pBM2 = kBM;
		oldBMs = SelectObject(sdc, pBM2->hBitmap);
		SelectPalette(sdc, kPal, 0);
		RealizePalette(sdc);

		oldstretchs = SetStretchBltMode(sdc, STRETCH_DELETESCANS); 
		oldstretchd = SetStretchBltMode(ddc, STRETCH_DELETESCANS); 
 
		rect.left = (int)GETREGD(REG_19);
		rect.top  = (int)GETREGD(REG_18);
		dw1 = GETREGD(REG_11);	// dst width
		dw2 = GETREGD(REG_10);	// dst height
		rect.right= rect.left + (int)dw1;
		rect.bottom=rect.top  + (int)dw2;

		StretchBlt(PRThdc, rect.left, rect.top, (int)dw1, (int)dw2, sdc, (int)GETREGD(REG_17), (int)GETREGD(REG_16), (int)GETREGD(REG_13), (int)GETREGD(REG_12), SRCCOPY);

		SetStretchBltMode(sdc, oldstretchs);
		SetStretchBltMode(ddc, oldstretchd);
		SelectObject(sdc, oldBMs);
		return;
//**** SYS_PRINTLINES
// inputs:
//	R14 = offset in memblock[R95] of (x,y) DWORD coordinate array
//	R13 = RGB color of lines
//	R39 = # of points in the [R14] array
//	R95 = 0 for R0, 1 for R1, 2 for R2 memory block for R14 offset
	 case	SYS_PRINTLINES:
	 // WARN: PTR4
		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);

		hPen = CreatePen(PS_SOLID, 1, GETREGD(REG_13));
		holdPen = SelectObject(PRThdc, hPen);

		pI = (int *)malloc(2*GETREGW(REG_39)*sizeof(int));
		rect.left = rect.top = 32000;
		rect.right = rect.bottom = 0;
		for(i=0; i<(int)GETREGW(REG_39); i++) {
			pI[2*i] = (int)(long)(*((DWORD *)pMem));
			if( (int)(long)(*((DWORD *)pMem))<rect.left ) rect.left = (int)(long)(*((DWORD *)pMem));
			if( (int)(long)(*((DWORD *)pMem))>rect.right) rect.right= (int)(long)(*((DWORD *)pMem));
			pMem += 4;
			pI[2*i+1] = (int)(long)(*((DWORD *)pMem));
			if( (int)(long)(*((DWORD *)pMem))<rect.top ) rect.top   = (int)(long)(*((DWORD *)pMem));
			if( (int)(long)(*((DWORD *)pMem))>rect.bottom) rect.bottom = (int)(long)(*((DWORD *)pMem));
			pMem += 4;
		}
		Polyline(PRThdc, (POINT *)pI, (int)GETREGW(REG_39));
		free(pI);
	
		SelectObject(PRThdc, holdPen);
		DeleteObject(hPen);
		return;
//**** SYS_PRINTGETTEXTSIZE
// assumes a SYS_PRINTSTART and SYS_BEGTEXT have already been done
// inputs:
//	R14 = string buffer address in memblk[R95]
//	R95 = memblk for string buffer
//	R37 = font handle
// outputs:
//	R15 = width
//	R14 = height
//	R13 = height of non-descending portion of font
	 case SYS_PRINTGETTEXTSIZE:
		w = GETREGW(REG_37);
		if( w<MAXFONTS && Fonts[w].used ) SelectObject(PRThdc, Fonts[w].hFont);
		SetTextAlign(PRThdc, TA_LEFT | TA_TOP);

		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);
		GetTextMetrics(PRThdc, &tm);
		GetTextExtentPoint32(PRThdc, (char *)pMem, strlen((char *)pMem), &txtsize);
		SETREGD(REG_15, (DWORD)(long)txtsize.cx);	// width of font
		SETREGD(REG_14, (DWORD)(long)txtsize.cy);	// height of font
		SETREGD(REG_13, (DWORD)(long)tm.tmAscent);	// height of non-descending portion of font
		return;

//**** SYS_ALLOCMEM
// Inputs:
//	R15 = size of memory block to allocate
// Outputs:
//	R15 = "handle" of memory block allocated (must be non-zero for valid memory allocation)
//		this "handle" must be place in R1 or R2 to access the memory (using LD1/LD2 type instructions)
//		==NULL if error
	 case	SYS_ALLOCMEM:
		// WARN: PTR4
		SETREGD(REG_15, (DWORD)malloc((size_t)GETREGD(REG_15)));
		return;
//**** SYS_FREEMEM
// Inputs:
//	R15 = "handle" of memory block to free
	 case	SYS_FREEMEM:
		free((void *)GETREGD(REG_15));
		return;
//**** SYS_TIMEREG
// Outputs:
//	R39 = year (ie, 2001)
//	R38 = month (0-11, Jan=0, Feb=1, Mar=2, etc)
//	R37 = day (1-31)
//	R36 = hour (0-23)
//	R35 = minute (0-59)
//	R34 = second (0-59)
	 case	SYS_TIMEREG:
		curtime = time(NULL);
		ourtime = localtime(&curtime);
		SETREGW(REG_39, (WORD)ourtime->tm_year+1900);
		SETREGW(REG_38, (WORD)ourtime->tm_mon);
		SETREGW(REG_37, (WORD)ourtime->tm_mday);
		SETREGW(REG_36, (WORD)ourtime->tm_hour);
		SETREGW(REG_35, (WORD)ourtime->tm_min);
		SETREGW(REG_34, (WORD)ourtime->tm_sec);
		return;
//**** SYS_TIMESTR
// Inputs:
//	R15 = offset in memblock[R95] of buffer for TIME string
//	R95 = 0 for R0, 1 for R1, 2 for R2
// Outputs:
//		buffer [R95]:R15 filled with string representing current time
	 case	SYS_TIMERSTR:
	 // WARN: PTR4
		pMem = (BYTE *)GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15);
		curtime = time(NULL);
		strcpy((char *)pMem, asctime(localtime(&curtime)));
		return;
//**** SYS_TIMEMILLI
// Outputs:
//	R15 = "progressive" time count in milliseconds.
//		By "progressive", I mean a count which doesn't necessarily bear
//		any resemblance to "real" time, but rather is a counter that is
//		a reasonably accurate "millisecond counter" that may eventually
//		roll over to 0, but which should not do so more often than once a day,
//		and who's count value continues upwards (until such time as a roll-over
//		might occur).  This is accomplished on the MS-Windows platform using
//		the GetTickCount() function.  Other platforms may choose to implement
//		this in whatever way seems easiest, while meeting the above criteria.
//		This counter value is only used for timing the speed of animations
//		and brief delays in actions.  So, reasonable accuracy down to 5 or 10
//		milliseconds would be very nice, and the roll-over (if and when it
//		occurs) should only cause (at most) a slight burp in any animation
//		or screen action that is going on at the moment (but from which it
//		should recover almost immediately).
//    NOTE: It is IMPORTANT that his count be relative accurate, as it is
//		used for generating the elapsed time count for solving puzzles!
	 case	SYS_TIMEMILLI:
		SETREGD(REG_15, GetTickCount());
		return;
//**** SYS_MEMCPY
// Inputs:
//	R15 = offset in memblk[R95] of destination buffer
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblk[R94] of source buffer
//	R94 = 0 for R0, 1 for R1, 2 for R2
//	R13 = number of bytes to copy
	 case	SYS_MEMCPY:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		memcpy(pMem, pMem2, (size_t)GETREGD(REG_13));
		return;
//**** SYS_MEMMOVE
// Inputs:
//	R15 = offset in memblk[R95] of destination buffer
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblk[R94] of source buffer
//	R94 = 0 for R0, 1 for R1, 2 for R2
//	R13 = number of bytes to move
	 case	SYS_MEMMOVE:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		memmove(pMem, pMem2, (size_t)GETREGD(REG_13));
		return;
//**** SYS_STRCPY
// Inputs:
//	R15 = offset in memblk[R95] of destination buffer
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblk[R94] of source buffer
//	R94 = 0 for R0, 1 for R1, 2 for R2
	 case	SYS_STRCPY:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		strcpy((char *)pMem, (char *)pMem2);
		return;
//**** SYS_STRCMP
// Inputs:
//	R15 = offset in memblk[R95] of buffer1
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblk[R94] of buffer2
//	R94 = 0 for R0, 1 for R1, 2 for R2
// Outputs:
//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
	 case	SYS_STRCMP:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		SETREGB(REG_95, (BYTE)(char)strcmp((char *)pMem, (char *)pMem2));
		return;
//**** SYS_STRICMP
// Inputs:
//	R15 = offset in memblk[R95] of buffer1
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblk[R94] of buffer2
//	R94 = 0 for R0, 1 for R1, 2 for R2
// Outputs:
//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
	 case	SYS_STRICMP:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		SETREGB(REG_95, (BYTE)(char)stricmp((char *)pMem, (char *)pMem2));
		return;
//**** SYS_STRLEN
// Inputs:
//	R15 = offset in memblk[R95] of buffer
//	R95 = 0 for R0, 1 for R1, 2 for R2
// Outputs:
//	R39 = length of nul-terminated string in buffer
	 case	SYS_STRLEN:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		SETREGW(REG_39, (WORD)strlen((char *)pMem));
		return;
//**** SYS_MEMSET
// Inputs:
//	R15 = offset in memblk[R95] of destination buffer
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R94 = value to write into bytes of buffer
//	R13 = number of bytes to set
	 case	SYS_MEMSET:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		memset(pMem, GETREGB(REG_94), (size_t)GETREGD(REG_13));
		return;
//**** SYS_MEMCMP
// Inputs:
//	R15 = offset in memblk[R95] of buffer1
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R14 = offset in memblk[R94] of buffer2
//	R94 = 0 for R0, 1 for R1, 2 for R2
//	R13 = number of bytes to compare
// Outputs:
//	R95 = <0 if buffer1<buffer2, 0 if buffer1==buffer2, >0 if buffer1>buffer2
//	E-flag set if R95==0, L-flag set if R95<0, G-flag set if R95>0
	 case	SYS_MEMCMP:
		pMem = (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_95))) + GETREGD(REG_15);
		pMem2= (BYTE *)(GETREGD(4*(DWORD)GETREGB(REG_94))) + GETREGD(REG_14);
		i = memcmp(pMem, pMem2, (size_t)GETREGD(REG_13));
		SETREGB(REG_95, (BYTE)((i==0)?0:((i & 0x8000000)?0xFF:0x01)));
		return;
//**** SYS_SETCURSOR
// Inputs:
//	R39 = 0 for normal, 1 for wait, 2 for hotlink
	 case SYS_SETCURSOR:
		CurCur = GETREGW(REG_39);
		// CurCur is used in the MainWnd WM_MOUSEMOVE code in MAIN.CPP to update the cursor
		if( CurCur==2 ) SetCursor(LoadCursor(NULL, IDC_UPARROW));
		else if( CurCur==1 ) SetCursor(LoadCursor(NULL, IDC_WAIT));
		else SetCursor(LoadCursor(NULL, IDC_ARROW));
		return;
//**** SYS_SETPGMTITLE
// Inputs:
//	R15 = offset in memblock[R95] of buffer containing program title
//	R95 = 0 for R0, 1 for R1, 2 for R2
	 case	SYS_SETPGMTITLE:
	 // WARN: PTR4
		pMem = (BYTE *)GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15);
		SetWindowText(hWndMain, (char *)pMem);
		return;
/*************************************************************
 * BITMAP functions: the KINT program creates a "display bitmap"
 * automatically.  A PDIBS handle of NULL refers to this
 * "display bitmap".  The NULL PDIBS cannot be deleted, but
 * it can be written to by all of the drawing functions.  The
 * KINT program is also responsible for any "cleanup" work that
 * needs to be done regarding the NULL "display bitmap" when
 * the program is exiting, as well as deleting all resources
 * allocated for other bitmaps created by SysCreBM, SysLoadBM,
 * and SysLoadMBM.
 *************************************************************/

//**** SYS_CREBM
// inputs:
//	R39 = width
//	R38 = height
// outputs:
//	R15 = handle (PDIBS in this implementation) of created bitmap, NULL if creation error.
	 case	SYS_CREBM:
		pBM = DibCreate(GETREGW(REG_39), GETREGW(REG_38));
		if( pBM!=NULL) {
			pBM->pNext = (void *)pFirstBM;
			if( pFirstBM!=NULL ) pFirstBM->pPrev = (void *)pBM;
			pBM->pPrev = NULL;
			pFirstBM = pBM;
		} else {
			MessageBox(NULL, "Error: SYS_CREBM", "Oops", MB_OK);
		}
	// WARN: PTR4
		SETREGD(REG_15, (DWORD)pBM);	// return to the Kode the pointer to the bitmap structure as the "bitmap handle"
		return;
//**** SYS_DELBM
// inputs:
//	R15 = handle (PDIBS in this implementation) of bitmap to delete
	 case	SYS_DELBM:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM!=NULL ) {
			if( pBM==pFirstBM ) {
				pFirstBM = (PDIBS)pBM->pNext;
				if( pFirstBM!=NULL ) pFirstBM->pPrev = NULL;
			} else {
				if( pBM->pNext!=NULL ) ((PDIBS)pBM->pNext)->pPrev = pBM->pPrev;
				if( pBM->pPrev!=NULL ) ((PDIBS)pBM->pPrev)->pNext = pBM->pNext;
			}
			DibDelete(pBM);
		}
		return;
//**** SYS_SIZEBM
// Inputs:
//	R15 = handle of bitmap
// Outputs:
//	R39 = width of bitmap
//	R38 = height of bitmap
	 case SYS_SIZEBM:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;
		SETREGW(REG_39, DibWidth(pBM));
		SETREGW(REG_38, DibHeight(pBM));
		return;
//**** SYS_DRAWBM
// Inputs:
//	R15 = handle of destination bitmap
//	R14 = handle of source bitmap
//	R39 = x-coord of destination
//	R38 = y-coord of destination
//	R37 = x-coord of source
//	R36 = y-coord of source
//	R35 = width of source
//	R34 = height of source
	 case	SYS_DRAWBM:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;
		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);

		pBM2 = (PDIBS)GETREGD(REG_14);
		if( pBM2==NULL ) pBM2 = kBM;
		oldBMs = SelectObject(sdc, pBM2->hBitmap);
		SelectPalette(sdc, kPal, 0);
		RealizePalette(sdc);

		rect.left = GETREGW(REG_39);
		rect.top  = GETREGW(REG_38);
		w = GETREGW(REG_35);
		h = GETREGW(REG_34);
		rect.right= rect.left + w;
		rect.bottom=rect.top  + h;

		BitBlt(ddc, rect.left, rect.top, w, h, sdc, GETREGW(REG_37), GETREGW(REG_36), SRCCOPY);
		if( pBM==kBM ) InvalidateRect(hWndMain, &rect, FALSE);
		SelectObject(ddc, oldBMd);
		SelectObject(sdc, oldBMs);
		return;
//**** SYS_STRETCHBM
// Inputs:
//	R15 = handle of destination bitmap
//	R14 = handle of source bitmap
//	R39 = x-coord of destination
//	R38 = y-coord of destination
//	R37 = x-coord of source
//	R36 = y-coord of source
//	R35 = width of source
//	R34 = height of source
//	R33 = width of destination
//	R32 = height of destination
	 case	SYS_STRETCHBM:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;
		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);

		pBM2 = (PDIBS)GETREGD(REG_14);
		if( pBM2==NULL ) pBM2 = kBM;
		oldBMs = SelectObject(sdc, pBM2->hBitmap);
		SelectPalette(sdc, kPal, 0);
		RealizePalette(sdc);

		oldstretchs = SetStretchBltMode(sdc, STRETCH_DELETESCANS); 
		oldstretchd = SetStretchBltMode(ddc, STRETCH_DELETESCANS); 
 
		rect.left = GETREGW(REG_39);
		rect.top  = GETREGW(REG_38);
		w = GETREGW(REG_33);
		h = GETREGW(REG_32);
		rect.right= rect.left + w;
		rect.bottom=rect.top  + h;

		StretchBlt(ddc, rect.left, rect.top, w, h, sdc, GETREGW(REG_37), GETREGW(REG_36), GETREGW(REG_35), GETREGW(REG_34), SRCCOPY);
		if( pBM==kBM ) InvalidateRect(hWndMain, &rect, FALSE);

		SetStretchBltMode(sdc, oldstretchs);
		SetStretchBltMode(ddc, oldstretchd);
		SelectObject(ddc, oldBMd);
		SelectObject(sdc, oldBMs);
		return;
//**** SYS_DRAWMBM
// Inputs:
//	R15 = handle of destination bitmap
//	R14 = handle of source bitmap
//	R39 = x-coord of destination
//	R38 = y-coord of destination
//	R37 = x-coord of source
//	R36 = y-coord of source
//	R35 = width of source
//	R34 = height of source
	 case	SYS_DRAWMBM:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;
		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);

		pBM2 = (PDIBS)GETREGD(REG_14);
		if( pBM2==NULL ) pBM2 = kBM;

		rect.left = GETREGW(REG_39);
		rect.top  = GETREGW(REG_38);
		w = GETREGW(REG_35);
		h = GETREGW(REG_34);
		rect.right= rect.left + w;
		rect.bottom=rect.top  + h;

		if( pBM2->hBitmapAnd != NULL ) {
			oldBMs = SelectObject(sdc, pBM2->hBitmapAnd);
		SelectPalette(sdc, kPal, 0);
		RealizePalette(sdc);
			BitBlt(ddc, rect.left, rect.top, w, h, sdc, GETREGW(REG_37), GETREGW(REG_36), SRCAND);
			SelectObject(sdc, oldBMs);
		}
		oldBMs = SelectObject(sdc, pBM2->hBitmap);
		SelectPalette(sdc, kPal, 0);
		RealizePalette(sdc);
		BitBlt(ddc, rect.left, rect.top, w, h, sdc, GETREGW(REG_37), GETREGW(REG_36), SRCPAINT);

		if( pBM==kBM ) InvalidateRect(hWndMain, &rect, FALSE);

		SelectObject(ddc, oldBMd);
		SelectObject(sdc, oldBMs);
		return;
//**** SYS_LOADBM
// Inputs:
//	R15 = offset into memory block[R95] of filename
//	R95 = 0 for R0, 1 for R1, 2 for R2
//	R94 =
//		0 file is in GAME folder
//		1 file is in UI folder
//		2 file is in IMAGES folder
//		3 file is in BACKGNDS folder
// Outputs:
//	R15 = handle (PDIBS) of loaded bitmap, or NULL if failed
	 case	SYS_LOADBM:
	 // WARN: PTR4
		pMem = (BYTE *)GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15);
		BuildPath(pMem, GETREGB(REG_94));
// BuildPath puts .BMP on for UI, IMAGES, BACKGNDS folders, but not GAME folder.
		pBM = DibFileLoad(TmpPath);
		if( pBM==NULL && GETREGB(REG_94)==0 ) {
			strcat(TmpPath, ".BMP");
			pBM = DibFileLoad(TmpPath);
		}
		if( pBM!=NULL ) {
			pBM->pNext = (void *)pFirstBM;
			if( pFirstBM!=NULL ) pFirstBM->pPrev = (void *)pBM;
			pBM->pPrev = NULL;
			pFirstBM = pBM;
		}
		SETREGD(REG_15, (DWORD)pBM);
		return;
//**** SYS_LOADMBM
// Inputs:
//	R15 = offset into memory block[R95] of filename
//	R39 = x-coord of transparent pixel
//	R38 = y-coord of transparent pixel
//	R95 = 0 for R0, 1 for R1, 2 for R2
//		NOTE: filename is relative to program's directory.  See SYS_FOPEN note.
// Outputs:
//	R15 = handle (PDIBS) of loaded bitmap, or NULL if failed
	 case	SYS_LOADMBM:
	 // WARN: PTR4
		pMem = (BYTE *)GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15);
		BuildPath(pMem, GETREGB(REG_94));
// BuildPath puts .BMP on for UI, IMAGES, BACKGNDS folders, but not GAME folder.
		pBM = DibFileLoad(TmpPath);
		if( pBM==NULL && GETREGB(REG_94)==0 ) {
			strcat(TmpPath, ".BMP");
			pBM = DibFileLoad(TmpPath);
		}
		SETREGD(REG_15, (DWORD)pBM);
		if( pBM!=NULL ) {
			COLORREF	oldBkColor, trans;

			int			w, h;
			HDC			invandDC;
			HBITMAP		invandBM, oldBMinv;
/*	Old Win 95 systems seem to be unable to properly deal with
	DIBsections when converting from 24-bit color to monochrome
	(for creating the mask).  So, we first create a DDB (device-
	dependant-bitmap), SRCCOPY the DIBsection into the DDB, then
	use the DDB to create the mask.  Hopefully other OS's aren't
	quite so brain-dead and this process will be easier...
*/
			HBITMAP		ddb, oldddb;
			HDC			ddbhdc, scrDC;
// link this new bitmap into the linked list for cleanup at program termination
			pBM->pNext = (void *)pFirstBM;
			if( pFirstBM!=NULL ) pFirstBM->pPrev = (void *)pBM;
			pBM->pPrev = NULL;
			pFirstBM = pBM;

			w = DibWidth(pBM);
			h = DibHeight(pBM);
			oldBMs = SelectObject(sdc, pBM->hBitmap);
		SelectPalette(sdc, kPal, 0);
		RealizePalette(sdc);
		// create the DDB DC and bitmap	and copy the DIB to it
			scrDC = GetDC(NULL);
			ddbhdc = CreateCompatibleDC(scrDC);
			ddb = CreateCompatibleBitmap(scrDC, w, h);
			ReleaseDC(NULL, scrDC);

			oldddb = SelectObject(ddbhdc, ddb);
		SelectPalette(ddbhdc, kPal, 0);
		RealizePalette(ddbhdc);
			BitBlt(ddbhdc, 0, 0, w, h, sdc, 0, 0, SRCCOPY);

		// Get color of pixel for TRANSPARENT COLOR
			trans = GetPixel(ddbhdc, (int)GETREGW(REG_39), (int)GETREGW(REG_38));
			oldBkColor = SetBkColor(ddbhdc, trans);
		// Create the monochrome AND mask
			pBM->hBitmapAnd = CreateBitmap(w, h, 1, 1, NULL);
			oldBMd = SelectObject(ddc, pBM->hBitmapAnd);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
			BitBlt(ddc, 0, 0, w, h, ddbhdc, 0, 0, SRCCOPY);
			SetBkColor(ddbhdc, oldBkColor);

		// Create an INVERSE AND mask from the AND mask
			invandDC = CreateCompatibleDC(ddc);
			invandBM = CreateBitmap(w, h, 1, 1, NULL);
			oldBMinv = SelectObject(invandDC, invandBM);
		SelectPalette(invandDC, kPal, 0);
		RealizePalette(invandDC);
			BitBlt(invandDC, 0, 0, w, h, ddc, 0, 0, NOTSRCCOPY);

		// Apply the INVERSE AND mask to the original image to create the OR mask
			BitBlt(sdc, 0, 0, w, h, invandDC, 0, 0, SRCAND);

			SelectObject(ddbhdc, oldddb);
			DeleteDC(ddbhdc);
			DeleteObject(ddb);

			SelectObject(invandDC, oldBMinv);
			DeleteDC(invandDC);
			DeleteObject(invandBM);

			SelectObject(sdc, oldBMs);
			SelectObject(ddc, oldBMd);
		}
		return;
//**** SYS_GETWINDOWPOS
// outputs:
//	R15 = X
//	R14 = Y
//	R13 = W
//	R12 = H
//	R11 = system dependant
	 case	SYS_GETWINDOWPOS:
		WndPlace.length = sizeof(WINDOWPLACEMENT);
		GetWindowPlacement(hWndMain, &WndPlace);
		SETREGD(REG_15, WndPlace.rcNormalPosition.left);
		SETREGD(REG_14, WndPlace.rcNormalPosition.top);
		SETREGD(REG_13, WndPlace.rcNormalPosition.right-WndPlace.rcNormalPosition.left);
		SETREGD(REG_12, WndPlace.rcNormalPosition.bottom-WndPlace.rcNormalPosition.top);
		showCmd = WndPlace.showCmd;
		if( showCmd==SW_MINIMIZE ) showCmd = SW_RESTORE;
		if( showCmd==SW_SHOWMINIMIZED ) showCmd = SW_SHOWNORMAL;
		SETREGD(REG_11, showCmd);
		return;
//**** SYS_SETWINDOWPOS
// inputs:
//	R15 = X
//	R14 = Y
//	R13 = W
//	R12 = H
//	R11 = system dependant
	 case	SYS_SETWINDOWPOS:
		WndPlace.length = sizeof(WINDOWPLACEMENT);
		WndPlace.flags = 0;
		mainShowCmd = WndPlace.showCmd = (UINT)GETREGD(REG_11);
		WndPlace.rcNormalPosition.left = (int)GETREGD(REG_15);
		WndPlace.rcNormalPosition.top  = (int)GETREGD(REG_14);
		WndPlace.rcNormalPosition.right= (int)GETREGD(REG_13)+(int)GETREGD(REG_15);
		WndPlace.rcNormalPosition.bottom=(int)GETREGD(REG_12)+(int)GETREGD(REG_14);
		WndPlace.ptMaxPosition.x = 0;
		WndPlace.ptMaxPosition.y = 0;
		SetWindowPlacement(hWndMain, &WndPlace);
		return;
//**** SYS_GETVIEWSIZE
// outputs:
//	R39 = window width
//	R38 = window height
	 case	SYS_GETVIEWSIZE:
		GetClientRect(hWndMain, &rect);
		SETREGW(REG_39, (WORD)rect.right);
		SETREGW(REG_38, (WORD)rect.bottom);
		return;
//**** SYS_GETSCREENSIZE
// outputs:
//	R39 = screen width
//	R38 = screen height
	 case	SYS_GETSCREENSIZE:
		SETREGW(REG_39, ScreenWidth);
		SETREGW(REG_38, ScreenHeight);
		SETREGW(REG_37, ScreenWidthIndex);
		return;
//**** SYS_FLUSHSCREEN
// force display update/refresh RIGHT NOW (rather than waiting a later "update message" or whatever)
	 case	SYS_FLUSHSCREEN:
		UpdateWindow(hWndMain);
		return;
//**** SYS_RECTEDGED:
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R14 = RGB color fill
//	R13 = RGB color frame
//	R39 = left
//	R38 = top
//	R37 = right
//	R36 = bottom
	 case	SYS_RECTEDGED:
// WARN: PTR4
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);

		hPen = CreatePen(PS_SOLID, 1, GETREGD(REG_13));
		hBrush = CreateSolidBrush(GETREGD(REG_14));
		rect.left = (int)((short *)GETREGW(REG_39));
		rect.top  = (int)((short *)GETREGW(REG_38));
		rect.right= (int)((short *)GETREGW(REG_37))+1;
		rect.bottom=(int)((short *)GETREGW(REG_36))+1;
		holdPen = SelectObject(ddc, hPen);
		holdBrush = SelectObject(ddc, hBrush);

		Rectangle(ddc, rect.left, rect.top, rect.right, rect.bottom);

		SelectObject(ddc, oldBMd);
		SelectObject(ddc, holdPen);
		SelectObject(ddc, holdBrush);
		DeleteObject(hPen);
		DeleteObject(hBrush);
		if( pBM==kBM ) InvalidateRect(hWndMain, &rect, FALSE);
		return;
//**** SYS_RECTSOLID:
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R14 = RGB color fill
//	R39 = left
//	R38 = top
//	R37 = right
//	R36 = bottom
	 case	SYS_RECTSOLID:
// WARN: PTR4
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
		
		hBrush = CreateSolidBrush((COLORREF)GETREGD(REG_14));
		rect.left = (int)((short *)GETREGW(REG_39));
		rect.top  = (int)((short *)GETREGW(REG_38));
		rect.right= (int)((short *)GETREGW(REG_37))+1;
		rect.bottom=(int)((short *)GETREGW(REG_36))+1;
		FillRect(ddc, &rect, hBrush);

		SelectObject(ddc, oldBMd);
		DeleteObject(hBrush);
		if( pBM==kBM ) InvalidateRect(hWndMain, &rect, FALSE);
		return;
//**** SYS_RECTOUTLINE:
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R13 = RGB color frame
//	R39 = left
//	R38 = top
//	R37 = right
//	R36 = bottom
	 case	SYS_RECTOUTLINE:
// WARN: PTR4
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);

		hPen = CreatePen(PS_SOLID, 1, GETREGD(REG_13));
		rect.left = pts[0].x = pts[3].x = pts[4].x = (int)((short *)GETREGW(REG_39));
		rect.top  = pts[0].y = pts[1].y = pts[4].y = (int)((short *)GETREGW(REG_38));
		rect.right= pts[1].x = pts[2].x = (int)((short *)GETREGW(REG_37));
		rect.bottom=pts[2].y = pts[3].y = (int)((short *)GETREGW(REG_36));
		holdPen = SelectObject(ddc, hPen);

		Polyline(ddc, pts, 5);

		SelectObject(ddc, oldBMd);
		SelectObject(ddc, holdPen);
		DeleteObject(hPen);
		if( pBM==kBM ) {
			++rect.right;
			++rect.bottom;
			InvalidateRect(hWndMain, &rect, FALSE);
		}
		return;
//**** SYS_LINES
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R14 = offset in memblock[R95] of (x,y) WORD coordinate array
//	R13 = RGB color of lines
//	R39 = # of points in the [R14] array
//	R95 = 0 for R0, 1 for R1, 2 for R2 memory block for R14 offset
	 case	SYS_LINES:
	 // WARN: PTR4
		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);

		hPen = CreatePen(PS_SOLID, 1, GETREGD(REG_13));
		holdPen = SelectObject(ddc, hPen);

		pI = (int *)malloc(2*GETREGW(REG_39)*sizeof(int));
		rect.left = rect.top = 32000;
		rect.right = rect.bottom = 0;
		for(i=0; i<(int)GETREGW(REG_39); i++) {
			pI[2*i] = (int)(short)(*((WORD *)pMem));
			if( (int)(*((WORD *)pMem))<rect.left ) rect.left = (int)(short)(*((WORD *)pMem));
			if( (int)(*((WORD *)pMem))>rect.right) rect.right= (int)(short)(*((WORD *)pMem));
			pMem += 2;
			pI[2*i+1] = (int)(short)(*((WORD *)pMem));
			if( (int)(*((WORD *)pMem))<rect.top ) rect.top   = (int)(short)(*((WORD *)pMem));
			if( (int)(*((WORD *)pMem))>rect.bottom) rect.bottom = (int)(short)(*((WORD *)pMem));
			pMem += 2;
		}
		Polyline(ddc, (POINT *)pI, (int)GETREGW(REG_39));
		free(pI);
	
		SelectObject(ddc, oldBMd);
		SelectObject(ddc, holdPen);
		DeleteObject(hPen);
		if( pBM==kBM ) {
			++rect.right;
			++rect.bottom;
			InvalidateRect(hWndMain, &rect, FALSE);
		}
		return;
//**** SYS_SETPIXEL:
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R14 = RGB color of point
//	R39 = x-coordinate of point
//	R38 = y-coordinate of point
	 case	SYS_SETPIXEL:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
		SetPixel(ddc, rect.left=(int)((short)GETREGW(REG_39)), rect.top=(int)((short)GETREGW(REG_38)), (COLORREF)GETREGD(REG_14));
		SelectObject(ddc, oldBMd);
		rect.right = rect.left+1;
		rect.bottom= rect.top+1;
		if( pBM==kBM ) InvalidateRect(hWndMain, &rect, FALSE);
		return;
//**** SYS_GETPIXEL:
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R39 = x-coordinate of point
//	R38 = y-coordinate of point
// outputs:
//	R14 = RGB color of point
	 case	SYS_GETPIXEL:
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
		SETREGD(REG_14, GetPixel(ddc, (int)((short)GETREGW(REG_39)), (int)((short)GETREGW(REG_38))));
		SelectObject(ddc, oldBMd);
		return;
//**** SYS_GETTEXTSIZE
// assumes a SYS_BEGTEXT and a SYS_TEXTSTYLE have already been done
// inputs:
//	R14 = string buffer address in memblk[R95]
//	R95 = memblk for string buffer
// outputs:
//	R39 = width
//	R38 = height
	 case SYS_GETTEXTSIZE:
		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);
		GetTextMetrics(ddc, &tm);
		GetTextExtentPoint32(ddc, (char *)pMem, strlen((char *)pMem), &txtsize);
		SETREGW(REG_39, txtsize.cx);	// width of font
		SETREGW(REG_38, txtsize.cy);	// height of font
		SETREGW(REG_37, tm.tmAscent);	// height of non-descending portion of font
		return;
//**** SYS_GETFONT
// inputs:
//	R14 = string buffer address in memblk[R95]
//	R95 = memblk for string buffer
//	R39 = pixel width of rectangle to fit text into
//	R38 = pixel height of rectangle to fit text into
//	R37 =
//		BIT 0 =0 for variable pitch, =1 for fixed pitch
//		BIT 1 =0 for normal, =1 for BOLD
//		BIT 2 =0 for normal, =1 for ITALIC
// outputs:
//	R39 = handle of font
//	R38 = height of font
	 case	SYS_GETFONT:
		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);
		w = FitFont((char *)pMem, GETREGW(REG_39), GETREGW(REG_38), GETREGW(REG_37), &txtsize);
		hfo = SelectObject(ddc, Fonts[w].hFont);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
		GetTextMetrics(ddc, &tm);
		SelectObject(ddc, hfo);
		SETREGW(REG_39, w);
		SETREGW(REG_38, txtsize.cy);
		SETREGW(REG_37, tm.tmAscent);	// height of non-descending portion of font
		return;
//**** SYS_RELFONT
// inputs:
//	R39 = handle of font to release
	 case	SYS_RELFONT:
		w = GETREGW(REG_39);
		if( w<MAXFONTS && Fonts[w].used ) {
			DeleteObject(Fonts[w].hFont);
			Fonts[w].used = 0;
		}
		return;
//**** SYS_BEGTEXT
// save whatever states you need to save regarding text output for "ddc"
// so that they can be restored at SYS_ENDTEXT
// inputs:
	 case	SYS_BEGTEXT:
		Moldie[MoldCnt].BkColor = GetBkColor(ddc);
		Moldie[MoldCnt].TextColor = GetTextColor(ddc);
		Moldie[MoldCnt].BkMode = GetBkMode(ddc);
		Moldie[MoldCnt].Font = SelectObject(ddc, Fonts[defFont].hFont);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
		++MoldCnt;
		return;
//**** SYS_ENDTEXT
// restore whatever states need to be restored that were saved at SYS_BEGTEXT
	 case	SYS_ENDTEXT:
		--MoldCnt;
		SetBkColor(ddc, Moldie[MoldCnt].BkColor);
		SetTextColor(ddc, Moldie[MoldCnt].TextColor);
		SetBkMode(ddc, Moldie[MoldCnt].BkMode);
		SelectObject(ddc, Moldie[MoldCnt].Font);
		return;
//**** SYS_TEXTSTYLE
//TextStyle(backcolor=-1 if transparent, textcolor, font, align)
// inputs:
//	R14 = backcolor (-1 if transparent, otherwise opaque)
//	R13 = textcolor
//	R39 = font handle
	 case	SYS_TEXTSTYLE:
		dw1 = GETREGD(REG_14);
		if( dw1==-1 ) SetBkMode(ddc, TRANSPARENT);
		else {
			SetBkMode(ddc, OPAQUE);
			SetBkColor(ddc, (COLORREF)dw1);
		}
		SetTextColor(ddc, (COLORREF)GETREGD(REG_13));

		w = GETREGW(REG_39);
		if( w<MAXFONTS && Fonts[w].used ) SelectObject(ddc, Fonts[w].hFont);
		SetTextAlign(ddc, TA_LEFT | TA_TOP);
		return;
//**** SYS_TEXT
// inputs:
//	R15 = PDIBS handle (NULL==screen)
//	R14 = string buffer address in memblk[R95]
//	R95 = memblk for string buffer
//	R39 = x-coordinate of point
//	R38 = y-coordinate of point
	 case	SYS_TEXT:
		pMem = (BYTE *)GETREGD(4*GETREGB(REG_95)) + GETREGD(REG_14);
		pBM = (PDIBS)GETREGD(REG_15);
		if( pBM==NULL ) pBM = kBM;

		oldBMd = SelectObject(ddc, pBM->hBitmap);
		SelectPalette(ddc, kPal, 0);
		RealizePalette(ddc);
		TextOut(ddc, (int)((short)GETREGW(REG_39)), (int)((short)GETREGW(REG_38)), (char *)pMem, strlen((char *)pMem));
		SelectObject(ddc, oldBMd);
		return;
//**** SYS_PLAYSOUND
// Inputs:
//	R15 = offset to memblock[R95] memory containing filename of sound to play
//	R95 = 0 if R0, 1 if R1, 2 if R2
	 case	SYS_PLAYSOUND:
		pMem = (BYTE *)GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15);
		strcpy(TmpPath, PgmPath);
	// note: on this platform, the .WAV files are in the 'home' folder
		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".WAV");
		sndPlaySound(TmpPath, SND_ASYNC | SND_NODEFAULT);
		return;
//**** SYS_PLAYMUSIC
// Inputs:
//	R15 = offset to memblock[R95] memory containing filename of music to play
//	R95 = 0 if R0, 1 if R1, 2 if R2
	 case	SYS_PLAYMUSIC:
		pMem = (BYTE *)GETREGD(4*(DWORD)GETREGB(REG_95)) + GETREGD(REG_15);
		strcpy(TmpPath, PgmPath);
		strcat(TmpPath, "..\\MUSIC\\");

		strcat(TmpPath, (char *)pMem);
		strcat(TmpPath, ".MID");
		playMIDIFile(hWndMain, TmpPath);
		return;
//**** SYS_STOPMUSIC
	 case	SYS_STOPMUSIC:
		if( MusicPlaying ) stopMIDIFile();
		return;
//**** SYS_ISMUSIC
// outputs:
//	R39=0 if no music playing, !=0 if music IS playing
	 case	SYS_ISMUSIC:
		SETREGW(REG_39, MusicPlaying);
		return;
//**** SYS_STARTTIMER
// inputs:
//	R15 = timer period in milliseconds
	 case	SYS_STARTTIMER:
	// create periodic timer
		if( TimerID==NULL ) TimerID = SetTimer(hWndMain, 1, (UINT)GETREGD(REG_15), NULL);
		return;
//**** SYS_STOPTIMER
	 case	SYS_STOPTIMER:
		if( TimerID!=NULL ) KillTimer(hWndMain, TimerID);
		TimerID = NULL;
		return;
//**** SYS_GETEOL
// inputs:
//	R15 = address in memblk R0 of where to push the OS's EOL sequence.
//		On MS-Windows, this pushes out a 13 and a 10.
//		On Linux and other OS's, it probably will push out ONLY a 10.
	 case SYS_GETEOL:
// for MS-Windows:
		SETMEMW((BYTE *)(GETREGD(REG_0)) + GETREGD(REG_15), 0x0A0D);
		SETREGD(REG_15, GETREGD(REG_15)+2);
// for unix-like OS's:
//		SETMEMB((BYTE *)(GETREGD(REG_0)) + GETREGD(REG_15), 0x0A);
//		SETREGD(REG_15, GETREGD(REG_15)+1);
		return;
	 case SYS_SYSTEMMSG:
		pMem = (BYTE *)GETREGD(REG_0) + GETREGD(REG_15);
		MessageBox(NULL, (const char *)pMem, "DEBUG", MB_OK);
		return;
	}
}

//***********************************************************
//* SysCleanup() is called when KINT is about to exit.		*
//* It is responsible for deleting/freeing all resources	*
//* which were allocated by SYS_ opcodes.					*
//***********************************************************
void SysCleanup() {

	PDIBS	pBM, pBMnext;
	int		i;

	for(pBM=pFirstBM; pBM!=NULL; ) {
		pBMnext = (PDIBS)pBM->pNext;
		DibDelete(pBM);
		pBM = pBMnext;
	}
	for(i=0; i<MAXFONTS; i++) {
		if( Fonts[i].used ) DeleteObject(Fonts[i].hFont);
	}

	DestroyPalette();
}

//********************
//* BITMAP FUNCTIONS *
//********************
//****************************************************************************
//* NOTE: much of this DIB code was shamelessly cribbed from Charles Petzold's
//* "Programming Windows" Fifth Edition.  I hacked and slashed it, and all
//* remaining bugs and deficiencies remain my responsibility.  Thanks, Charles!

/*----------------------------------------------------------------
   DibNumColors:  Returns the number of colors in the color table
  ----------------------------------------------------------------*/
int DibNumColors (PDIBS pdib)
{
     if( pdib->ds.dsBmih.biClrUsed != 0 ) return pdib->ds.dsBmih.biClrUsed;
     else if( DibBitCount(pdib) <= 8 ) return 1 << DibBitCount(pdib);
     return 0 ;
}

//********************************************************
//* Obtain entry from the DIB color table
//*
BOOL DibGetColor (PDIBS pdib, int index, RGBQUAD * prgb) {

     HDC        hdcMem ;
     int        iReturn ;

     hdcMem = CreateCompatibleDC(NULL);
     SelectObject(hdcMem, pdib->hBitmap);
		SelectPalette(hdcMem, kPal, 0);
		RealizePalette(hdcMem);
     iReturn = GetDIBColorTable(hdcMem, index, 1, prgb);
     DeleteDC(hdcMem);

     return iReturn ? TRUE : FALSE ;
}

//******************************************
DWORD DibBitsSize(PDIBS hdib) {

	PDIBS	pdib = (PDIBS)hdib;

	if( pdib->ds.dsBmih.biSizeImage!=0 ) return pdib->ds.dsBmih.biSizeImage;
	return DibHeight(hdib)*DibRowLength(hdib);
}

/*--------------------------------------------------------------
   Calculating shift values from color masks is required by the 
     DibCreateFromInfo function.
  --------------------------------------------------------------*/
int MaskToRShift (DWORD dwMask) {

	int		iShift;

	if( dwMask==0 ) return 0 ;
	for(iShift=0; !(dwMask & 1); iShift++) dwMask >>= 1 ;
	return iShift ;
}

int MaskToLShift (DWORD dwMask) {

	int iShift ;

	if( dwMask==0 ) return 0 ;
	while( !(dwMask & 1) ) dwMask >>= 1 ;
	for(iShift=0; dwMask & 1; iShift++) dwMask >>= 1 ;
	return 8-iShift;
}

/*-------------------------------------------------------------------------
   DibCreateFromInfo: All DIB creation functions ultimately call this one.
     This function is responsible for calling CreateDIBSection, allocating
     memory for DIBS, and setting up the row pointer.
  -------------------------------------------------------------------------*/
PDIBS DibCreateFromInfo(BITMAPINFO *pbmi) {

	BYTE	*pBits;
	DIBS	*pdib;
	HBITMAP	hBitmap;
	int		i;

	hBitmap = CreateDIBSection(NULL, pbmi, /*DIB_PAL_COLORS*/ DIB_RGB_COLORS, (void **)&pBits, NULL, 0);

	if( hBitmap==NULL ) return NULL;

	if( NULL==(pdib=(PDIBS)malloc(sizeof (DIBS)))) {
		DeleteObject(hBitmap);
		return NULL;
	}

	pdib->hBitmap = hBitmap;
	pdib->hBitmapAnd = NULL;
	pdib->pBits = pBits;

	GetObject(hBitmap, sizeof(DIBSECTION), &pdib->ds);

	if( DibCompression(pdib)==BI_BITFIELDS ) {
	// If the compression is BI_BITFIELDS, calculate shifts from masks
		for(i=0; i<3; i++) {
			pdib->iLShift[i] = MaskToLShift(pdib->ds.dsBitfields[i]);
			pdib->iRShift[i] = MaskToRShift(pdib->ds.dsBitfields[i]);
		}
	} else if( DibCompression(pdib)==BI_RGB ) {
	// If the compression is BI_RGB, but bit-count is 16 or 32,
	//   set the bitfields and the masks
		if( DibBitCount(pdib)==16 ) {
			pdib->ds.dsBitfields[0] = 0x00007C00;
			pdib->ds.dsBitfields[1] = 0x000003E0;
			pdib->ds.dsBitfields[2] = 0x0000001F;

			pdib->iRShift[0] = 10;
			pdib->iRShift[1] =  5;
			pdib->iRShift[2] =  0; 

			pdib->iLShift[0] =  3;
			pdib->iLShift[1] =  3;
			pdib->iLShift[2] =  3;
		} else if( DibBitCount(pdib)==24 || DibBitCount(pdib)==32 ) {
			pdib->ds.dsBitfields[0] = 0x00FF0000;
			pdib->ds.dsBitfields[1] = 0x0000FF00;
			pdib->ds.dsBitfields[2] = 0x000000FF;

			pdib->iRShift[0] = 16;
			pdib->iRShift[1] =  8;
			pdib->iRShift[2] =  0; 

			pdib->iLShift[0] =  0;
			pdib->iLShift[1] =  0;
			pdib->iLShift[2] =  0;
		}
	}
	return pdib;
}

/*-----------------------------------------------------
   DibFileLoad:  Creates a DIB section from a DIB file
  -----------------------------------------------------*/

PDIBS DibFileLoad(char *fname) {

	BITMAPFILEHEADER	bmfh;
	BITMAPINFO			*pbmi;
	DWORD				dwInfoSize, dwBitsSize;
	WORD				wBytesRead, wWant;
	PDIBS				pdib;
	BYTE				*pB;
	int					fh;

	// Open the file: read access, prohibit write access
	if( -1==(fh=_open(fname, _O_BINARY | _O_RDONLY)) ) return NULL;
	// Read in the BITMAPFILEHEADER
	wBytesRead = _read(fh, &bmfh, sizeof(BITMAPFILEHEADER));
	if( (wBytesRead != sizeof (BITMAPFILEHEADER)) || (bmfh.bfType != *(WORD *)"BM") ) {
		_close(fh);
		return NULL;
	}
	// Allocate memory for the information structure & read it in
	dwInfoSize = bmfh.bfOffBits - sizeof (BITMAPFILEHEADER);

	if( NULL==(pbmi=(struct tagBITMAPINFO *)malloc(dwInfoSize)) ) {
		_close(fh);
		return NULL;
	}
	wBytesRead = _read(fh, pbmi, dwInfoSize);
	if( wBytesRead != dwInfoSize )	{
		_close(fh);
		free(pbmi);
		return NULL;
	}
	  // Create the DIB
	pdib = DibCreateFromInfo(pbmi);
	free(pbmi);

	if( pdib==NULL ) {
		_close(fh);
		return NULL;
	}
	// Read in the bits
	dwBitsSize = bmfh.bfSize - bmfh.bfOffBits ;

	pB = (BYTE *)pdib->pBits;
	while( dwBitsSize ) {
		if( dwBitsSize>16384 ) wWant = 16384;
		else wWant = (WORD)dwBitsSize;
		if( wWant!=_read(fh, pB, wWant) ) {
			DibDelete(pdib);
			_close(fh);
			return NULL;
		}
		pB += wWant;
		dwBitsSize -= (DWORD)wWant;
	}
	_close(fh);
	return pdib;
}

//***********************************************
PDIBS DibCreate(DWORD w, DWORD h) {

	BITMAPINFO			pbmi;
	PDIBS				pdib;

	pbmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	pbmi.bmiHeader.biWidth = w;
	pbmi.bmiHeader.biHeight = h;
	pbmi.bmiHeader.biPlanes = 1;
	pbmi.bmiHeader.biBitCount = 24;
	pbmi.bmiHeader.biSizeImage = 4*((8*3*w+31)/32)*h;
	pbmi.bmiHeader.biXPelsPerMeter = pbmi.bmiHeader.biYPelsPerMeter = pbmi.bmiHeader.biClrUsed = pbmi.bmiHeader.biClrImportant = pbmi.bmiHeader.biCompression = 0;

	pdib = DibCreateFromInfo(&pbmi);

	return pdib;
}

//***********************************************
PDIBS DibCopy(PDIBS hdibSrc) {

	BITMAPINFO	*pbmi;
	PDIBS		hdibDst;
	int			i, iNumColors;
	RGBQUAD		*prgb;

	// Allocate the memory for the BITMAPINFO
	if( NULL==(pbmi=(struct tagBITMAPINFO *)malloc(DibInfoSize(hdibSrc))) ) return NULL;
	// Copy the information header
	CopyMemory(pbmi, DibInfoHeaderPtr(hdibSrc), sizeof(BITMAPINFOHEADER));
	// Copy the possible color masks
	prgb = (RGBQUAD *)((BYTE *)pbmi + sizeof(BITMAPINFOHEADER));
	if( DibMaskSize(hdibSrc) ) {
		CopyMemory(prgb, DibMaskPtr(hdibSrc), 3*sizeof(DWORD));
		prgb = (RGBQUAD *)((BYTE *)prgb + 3*sizeof(DWORD));
	}
	// Copy the color table
	iNumColors = DibNumColors(hdibSrc);
	for(i=0; i<iNumColors; i++) DibGetColor(hdibSrc, i, prgb+i);
	// Create the DIB copy
	hdibDst = DibCreateFromInfo(pbmi);
	// free the BITMAPINFO memory
	free(pbmi);

	if( hdibDst==NULL ) return NULL;
	CopyMemory((BYTE *)DibBitsPtr(hdibDst), (BYTE *)DibBitsPtr(hdibSrc), DibBitsSize(hdibSrc));
	return hdibDst;
}

//****************************************************
//* Free all memory for the DIB section
//***
BOOL DibDelete(PDIBS pdib) {

	DeleteObject(pdib->hBitmap);
	if( pdib->hBitmapAnd != NULL ) DeleteObject(pdib->hBitmapAnd);
	free(pdib) ;
	return TRUE;
}

#define	VARFONTNAME	"Arial"
//#define	VARFONTNAME	"Times New Roman"
//#define	VARFONTNAME	"MS Sans Serif"

//***************************************************************
// Creates a font such that "text" will be as large as possible
// while still fitting within a rectangle of width "w" and height "h".
WORD FitFont(char *text, int w, int h, WORD flags, SIZE *fontsize) {

	HDC		hdc;
	HFONT	hFont, hOldFont;
	SIZE	siz;
	int		i, len;

	hdc = GetDC(NULL);
	len = strlen(text);
	i = 8;
	do {
		++i;
		++i;
		if( flags & 1 ) hFont = CreateFont(-i, 0, 0, 0, (flags & 2)?FW_BOLD:FW_NORMAL, (flags & 4)?TRUE:FALSE, 0, 0, ANSI_CHARSET, OUT_STRING_PRECIS, 0, PROOF_QUALITY, FIXED_PITCH+FF_MODERN, "Courier New");
		else hFont = CreateFont(-i, 0, 0, 0, (flags & 2)?FW_BOLD:FW_NORMAL, (flags & 4)?TRUE:FALSE, 0, 0, ANSI_CHARSET, OUT_STRING_PRECIS, 0, PROOF_QUALITY, VARIABLE_PITCH+FF_MODERN, VARFONTNAME);
		hOldFont = (HFONT)SelectObject(hdc, hFont);
		GetTextExtentPoint32(hdc, text, len, &siz);
		SelectObject(hdc, hOldFont);
		DeleteObject(hFont);
		if( len==0 || siz.cx>=w || siz.cy>=h ) break;
	} while(TRUE);
	do {
		--i;
		--i;
		if( flags & 1 ) hFont = CreateFont(-i, 0, 0, 0, (flags & 2)?FW_BOLD:FW_NORMAL, (flags & 4)?TRUE:FALSE, 0, 0, ANSI_CHARSET, OUT_STRING_PRECIS, 0, PROOF_QUALITY, FIXED_PITCH+FF_MODERN, "Courier New");
		else hFont = CreateFont(-i, 0, 0, 0, (flags & 2)?FW_BOLD:FW_NORMAL, (flags & 4)?TRUE:FALSE, 0, 0, ANSI_CHARSET, OUT_STRING_PRECIS, 0, PROOF_QUALITY, VARIABLE_PITCH+FF_SWISS, VARFONTNAME);
		hOldFont = (HFONT)SelectObject(hdc, hFont);
		GetTextExtentPoint32(hdc, text, len, &siz);
		SelectObject(hdc, hOldFont);
		if( i<=4 || (siz.cx<w && siz.cy<h) ) break;
		DeleteObject(hFont);
	} while(TRUE);
	if( len==0 || siz.cx>=w || siz.cy>=h ) {
		DeleteObject(hFont);
		if( flags & 1 ) hFont = CreateFont(-8, 0, 0, 0, (flags & 2)?FW_BOLD:FW_NORMAL, (flags & 4)?TRUE:FALSE, 0, 0, ANSI_CHARSET, OUT_STRING_PRECIS, 0, PROOF_QUALITY, FIXED_PITCH+FF_MODERN, "Courier New");
		else hFont = CreateFont(-8, 0, 0, 0, (flags & 2)?FW_BOLD:FW_NORMAL, (flags & 4)?TRUE:FALSE, 0, 0, ANSI_CHARSET, OUT_STRING_PRECIS, 0, PROOF_QUALITY, VARIABLE_PITCH+FF_SWISS, VARFONTNAME);
	}

	if( fontsize!=NULL ) {
		hOldFont = (HFONT)SelectObject(hdc, hFont);
		GetTextExtentPoint32(hdc, "W", 1, fontsize);
		SelectObject(hdc, hOldFont);
	}
	ReleaseDC(NULL, hdc);

	for(i=0; i<MAXFONTS; i++) if( Fonts[i].used==0 ) break;
	if( i>=MAXFONTS ) {
		MessageBox(NULL, "Too many FONTS created by .KE", "Oops", MB_OK);
		DeleteObject(hFont);
	} else {
		Fonts[i].hFont = hFont;
		Fonts[i].used = 1;
	}
	return i;
}

//**************************************
void GetDefaultFont() {

	defFont = FitFont("Wy", ScreenWidth, ScreenHeight/40, 0, NULL);
}

//****************************************
void InitScreen() {

	HDC			hDC;
	HBITMAP		hbm;
	int			redraw;

	hDC = GetDC(NULL);
	ScreenWidth = GetDeviceCaps(hDC, HORZRES);
	if( ScreenWidth<800 ) ScreenWidthIndex = 0;
	else if( ScreenWidth<1024 ) ScreenWidthIndex = 1;
	else if( ScreenWidth<1152 ) ScreenWidthIndex = 2;
	else if( ScreenWidth<1280 ) ScreenWidthIndex = 3;
	else if( ScreenWidth<1600 ) ScreenWidthIndex = 4;
	else ScreenWidthIndex = 5;
	ScreenHeight = GetDeviceCaps(hDC, VERTRES);

	if( defFont!=-1 && Fonts[defFont].used ) {
		DeleteObject(Fonts[defFont].hFont);
		Fonts[defFont].used = 0;
	}
	GetDefaultFont();

	if( kBM!=NULL ) {
		DibDelete(kBM);
		redraw = 1;
	} else redraw = 0;
	kBM = DibCreate(ScreenWidth, ScreenHeight);
	if( NULL==kBM ) {
		MessageBox(NULL, "Failed creating screen BITMAP", "Oops", MB_OK);
		ReleaseDC(NULL, hDC);
		exit(0);
	}
	ReleaseDC(NULL, hDC);

	hbm = SelectObject(ddc, kBM->hBitmap);
	SelectObject(ddc, GetStockObject(BLACK_PEN));
	SelectObject(ddc, GetStockObject(BLACK_BRUSH));
	Rectangle(ddc, 0, 0, ScreenWidth, ScreenHeight);
	SelectObject(ddc, hbm);

	if( redraw ) {
// call the SCREEN_CHANGE code in the KE module
/****/	KEevent(KHDR_SCREEN_CHANGE, 0, 0, 0);
	}
}

//*****************************************************************
// The following palette is used for 256 color displays.
COLORREF	OurPalRGB[256];

BYTE	kLogPal[sizeof(LOGPALETTE)+(4096-1)*sizeof(PALETTEENTRY)];
BYTE	OurPal[sizeof(LOGPALETTE)+255*sizeof(PALETTEENTRY)] = {
			0, 3,	// version
			0, 1,	// number of entries
// 1 BLACK/WHITE
		0x00, 0x00, 0x00, 0x00,
		0x22, 0x22, 0x22, 0x00,
		0x33, 0x33, 0x33, 0x00,
		0x44, 0x44, 0x44, 0x00,
		0x55, 0x55, 0x55, 0x00,
		0x66, 0x66, 0x66, 0x00,
		0x77, 0x77, 0x77, 0x00,
		0x88, 0x88, 0x88, 0x00,
		0x99, 0x99, 0x99, 0x00,
		0xAA, 0xAA, 0xAA, 0x00,
		0xBB, 0xBB, 0xBB, 0x00,
		0xCC, 0xCC, 0xCC, 0x00,
		0xDD, 0xDD, 0xDD, 0x00,
		0xEE, 0xEE, 0xEE, 0x00,
		0xFF, 0xFF, 0xFF, 0x00,
// 2 BROWN
		0x28, 0x14, 0x00, 0x00,
		0x41, 0x23, 0x00, 0x00,
		0x5F, 0x32, 0x00, 0x00,
		0x82, 0x41, 0x00, 0x00,
		0xA0, 0x50, 0x00, 0x00,
		0xC3, 0x5F, 0x14, 0x00,
		0xE1, 0x73, 0x1E, 0x00,
		0xFF, 0x82, 0x32, 0x00,
		0xFF, 0x91, 0x41, 0x00,
		0xFF, 0xA0, 0x50, 0x00,
		0xFF, 0xAF, 0x5F, 0x00,
		0xFF, 0xBE, 0x73, 0x00,
		0xFF, 0xD2, 0x82, 0x00,
		0xFF, 0xE1, 0x91, 0x00,
		0xFF, 0xF0, 0xA0, 0x00,
// 3 BROWN
		0x32, 0x1E, 0x1E, 0x00,
		0x41, 0x20, 0x20, 0x00,
		0x5F, 0x30, 0x30, 0x00,
		0x82, 0x40, 0x40, 0x00,
		0xA0, 0x50, 0x50, 0x00,
		0xBE, 0x60, 0x60, 0x00,
		0xE1, 0x70, 0x70, 0x00,
		0xFF, 0x7F, 0x7F, 0x00,
		0xFF, 0x7F, 0x7F, 0x00,
		0xFF, 0x8E, 0x7F, 0x00,
		0xFF, 0x9F, 0x7F, 0x00,
		0xFF, 0xAF, 0x7F, 0x00,
		0xFF, 0xBF, 0x7F, 0x00,
		0xFF, 0xCF, 0x7F, 0x00,
		0xFF, 0xDF, 0x7F, 0x00,
// 4 BRICK RED
		0x28, 0x0D, 0x0D, 0x00,
		0x40, 0x15, 0x15, 0x00,
		0x60, 0x20, 0x20, 0x00,
		0x80, 0x2A, 0x2A, 0x00,
		0xA0, 0x35, 0x35, 0x00,
		0xC0, 0x40, 0x40, 0x00,
		0xE0, 0x4A, 0x4A, 0x00,

		0xFF, 0x55, 0x55, 0x00,
		0xFF, 0x67, 0x64, 0x00,
		0xFF, 0x6F, 0x74, 0x00,
		0xFF, 0x75, 0x84, 0x00,
		0xFF, 0x84, 0x9D, 0x00,
		0xFF, 0x94, 0xB7, 0x00,
		0xFF, 0x9F, 0xD1, 0x00,
		0xFF, 0xAE, 0xEA, 0x00,
// 5 RED TO ORANGE TO YELLOW
		0x90, 0x14, 0x00, 0x00,
		0xA0, 0x20, 0x00, 0x00,
		0xB0, 0x30, 0x00, 0x00,
		0xC0, 0x40, 0x00, 0x00,
		0xD0, 0x50, 0x00, 0x00,
		0xE0, 0x60, 0x00, 0x00,
		0xF0, 0x70, 0x00, 0x00,
		0xFF, 0x80, 0x00, 0x00,
		0xFF, 0x90, 0x00, 0x00,
		0xFF, 0xA0, 0x00, 0x00,
		0xFF, 0xB0, 0x00, 0x00,
		0xFF, 0xC0, 0x00, 0x00,
		0xFF, 0xD0, 0x00, 0x00,
		0xFF, 0xE0, 0x00, 0x00,
		0xFF, 0xF0, 0x00, 0x00,
// 6 RED
		0x28, 0x00, 0x00, 0x00,
		0x40, 0x00, 0x00, 0x00,
		0x60, 0x00, 0x00, 0x00,
		0x80, 0x00, 0x00, 0x00,
		0xA0, 0x00, 0x00, 0x00,
		0xC0, 0x00, 0x00, 0x00,
		0xE0, 0x00, 0x00, 0x00,
		0xFF, 0x00, 0x00, 0x00,
		0xFF, 0x28, 0x28, 0x00,
		0xFF, 0x40, 0x40, 0x00,
		0xFF, 0x60, 0x60, 0x00,
		0xFF, 0x80, 0x80, 0x00,
		0xFF, 0xA0, 0xA0, 0x00,
		0xFF, 0xC0, 0xC0, 0x00,
		0xFF, 0xE0, 0xE0, 0x00,
// 7 PURPLE
		0x28, 0x00, 0x28, 0x00,
		0x40, 0x00, 0x40, 0x00,
		0x60, 0x00, 0x60, 0x00,
		0x80, 0x00, 0x80, 0x00,
		0xA0, 0x00, 0xA0, 0x00,
		0xC0, 0x00, 0xC0, 0x00,
		0xE0, 0x00, 0xE0, 0x00,
		0xFF, 0x00, 0xFF, 0x00,
		0xFF, 0x28, 0xFF, 0x00,
		0xFF, 0x40, 0xFF, 0x00,
		0xFF, 0x60, 0xFF, 0x00,
		0xFF, 0x80, 0xFF, 0x00,
		0xFF, 0xA0, 0xFF, 0x00,
		0xFF, 0xC0, 0xFF, 0x00,
		0xFF, 0xE0, 0xFF, 0x00,
// 8 LIGHT PURPLE
		0x28, 0x14, 0x28, 0x00,
		0x40, 0x20, 0x40, 0x00,
		0x60, 0x30, 0x60, 0x00,
		0x80, 0x40, 0x80, 0x00,
		0xA0, 0x50, 0xA0, 0x00,
		0xC0, 0x60, 0xC0, 0x00,
		0xE0, 0x70, 0xE0, 0x00,
		0xFF, 0x7F, 0xFF, 0x00,
		0xFF, 0x94, 0xFF, 0x00,
		0xFF, 0xA0, 0xFF, 0x00,
		0xFF, 0xB0, 0xFF, 0x00,
		0xFF, 0xC0, 0xFF, 0x00,
		0xFF, 0xD0, 0xFF, 0x00,
		0xFF, 0xE0, 0xFF, 0x00,
		0xFF, 0xF0, 0xFF, 0x00,
// 9 DARK PURPLE
		0x28, 0x00, 0x50, 0x00,
		0x35, 0x05, 0x66, 0x00,
		0x42, 0x0A, 0x7C, 0x00,
		0x4F, 0x0F, 0x92, 0x00,
		0x5C, 0x14, 0xA8, 0x00,
		0x69, 0x19, 0xBE, 0x00,
		0x76, 0x1E, 0xD4, 0x00,
		0x83, 0x23, 0xEA, 0x00,
		0x90, 0x28, 0xFF, 0x00,
		0xA0, 0x40, 0xFF, 0x00,
		0xB0, 0x60, 0xFF, 0x00,
		0xC0, 0x80, 0xFF, 0x00,
		0xD0, 0xA0, 0xFF, 0x00,
		0xE0, 0xC0, 0xFF, 0x00,
		0xF0, 0xE0, 0xFF, 0x00,
// 10 BLUE
		0x00, 0x00, 0x28, 0x00,
		0x00, 0x00, 0x40, 0x00,
		0x00, 0x00, 0x60, 0x00,
		0x00, 0x00, 0x80, 0x00,
		0x00, 0x00, 0xA0, 0x00,
		0x00, 0x00, 0xC0, 0x00,
		0x00, 0x00, 0xE0, 0x00,
		0x00, 0x00, 0xFF, 0x00,
		0x0A, 0x28, 0xFF, 0x00,
		0x28, 0x4A, 0xFF, 0x00,
		0x46, 0x6A, 0xFF, 0x00,
		0x67, 0x8A, 0xFF, 0x00,
		0x87, 0xAA, 0xFF, 0x00,
		0xA7, 0xCA, 0xFF, 0x00,
		0xC7, 0xEB, 0xFF, 0x00,
// 11 DARK CYAN
		0x0F, 0x1E, 0x1E, 0x00,
		0x14, 0x23, 0x23, 0x00,
		0x19, 0x32, 0x32, 0x00,
		0x1E, 0x41, 0x41, 0x00,
		0x28, 0x50, 0x50, 0x00,
		0x32, 0x5F, 0x5F, 0x00,
		0x37, 0x73, 0x73, 0x00,
		0x41, 0x82, 0x82, 0x00,
		0x46, 0x91, 0x91, 0x00,
		0x50, 0xA0, 0xA0, 0x00,
		0x5A, 0xAF, 0xAF, 0x00,
		0x5F, 0xC3, 0xC3, 0x00,
		0x69, 0xD2, 0xD2, 0x00,
		0x73, 0xE1, 0xE1, 0x00,
		0x78, 0xF0, 0xF0, 0x00,
// 12 CYAN
		0x00, 0x28, 0x28, 0x00,
		0x00, 0x40, 0x40, 0x00,
		0x00, 0x60, 0x60, 0x00,
		0x00, 0x80, 0x80, 0x00,
		0x00, 0xA0, 0xA0, 0x00,
		0x00, 0xC0, 0xC0, 0x00,
		0x00, 0xE0, 0xE0, 0x00,
		0x00, 0xFF, 0xFF, 0x00,
		0x28, 0xFF, 0xFF, 0x00,
		0x40, 0xFF, 0xFF, 0x00,
		0x60, 0xFF, 0xFF, 0x00,
		0x80, 0xFF, 0xFF, 0x00,
		0xA0, 0xFF, 0xFF, 0x00,
		0xC0, 0xFF, 0xFF, 0x00,
		0xE0, 0xFF, 0xFF, 0x00,
// 13 GREEN
		0x00, 0x28, 0x00, 0x00,
		0x00, 0x40, 0x00, 0x00,
		0x00, 0x60, 0x00, 0x00,
		0x00, 0x80, 0x00, 0x00,
		0x00, 0xA0, 0x00, 0x00,
		0x00, 0xC0, 0x00, 0x00,
		0x00, 0xE0, 0x00, 0x00,
		0x00, 0xFF, 0x00, 0x00,
		0x28, 0xFF, 0x28, 0x00,
		0x40, 0xFF, 0x40, 0x00,
		0x60, 0xFF, 0x60, 0x00,
		0x80, 0xFF, 0x80, 0x00,
		0xA0, 0xFF, 0xA0, 0x00,
		0xC0, 0xFF, 0xC0, 0x00,
		0xE0, 0xFF, 0xE0, 0x00,
// 14 GREEN TO YELLOW
		0x0A, 0x23, 0x0A, 0x00,
		0x23, 0x41, 0x23, 0x00,
		0x32, 0x5F, 0x32, 0x00,
		0x41, 0x82, 0x41, 0x00,
		0x50, 0xA0, 0x50, 0x00,
		0x5F, 0xC3, 0x5F, 0x00,
		0x73, 0xE1, 0x73, 0x00,
		0x82, 0xFF, 0x82, 0x00,
		0x91, 0xFF, 0x6E, 0x00,
		0xA0, 0xFF, 0x5F, 0x00,
		0xB4, 0xFF, 0x50, 0x00,
		0xC3, 0xFF, 0x41, 0x00,
		0xD2, 0xFF, 0x32, 0x00,
		0xE1, 0xFF, 0x23, 0x00,
		0xF0, 0xFF, 0x0F, 0x00,
// 15 YELLOW
		0x28, 0x28, 0x00, 0x00,
		0x40, 0x40, 0x00, 0x00,
		0x60, 0x60, 0x00, 0x00,
		0x80, 0x80, 0x00, 0x00,
		0xA0, 0xA0, 0x00, 0x00,
		0xC0, 0xC0, 0x00, 0x00,
		0xE0, 0xE0, 0x00, 0x00,
		0xFF, 0xFF, 0x00, 0x00,
		0xFF, 0xFF, 0x28, 0x00,
		0xFF, 0xFF, 0x40, 0x00,
		0xFF, 0xFF, 0x60, 0x00,
		0xFF, 0xFF, 0x80, 0x00,
		0xFF, 0xFF, 0xA0, 0x00,
		0xFF, 0xFF, 0xC0, 0x00,
		0xFF, 0xFF, 0xE0, 0x00
	};

int		clrmap16[16]={0,16,32,48,64,80,104,120,136,152,168,184,200,224,240,255};

//***************************************************************************
void MakePalette() {

	int	r, g, b;
	HDC		hDC;
	
	hDC = GetDC(NULL);
	bitspixel = GetDeviceCaps(hDC, BITSPIXEL);
	planes = GetDeviceCaps(hDC, PLANES);
	numcolors = GetDeviceCaps(hDC, NUMCOLORS);
	sizepalette = GetDeviceCaps(hDC, SIZEPALETTE);
	ReleaseDC(NULL, hDC);

	if( bitspixel*planes==8 || sizepalette==256 ) colormode = 1;
	else if( bitspixel*planes>8 || numcolors>256 ) colormode = 2;
	else colormode = 0;

	if( colormode==2 ) {
		for(b=0; b<16; b++) {
			for(g=0; g<16; g++) {
				for(r=0; r<16; r++) {
					kLogPal[4+4*(256*b+16*g+r)+0] = clrmap16[r];
					kLogPal[4+4*(256*b+16*g+r)+1] = clrmap16[g];
					kLogPal[4+4*(256*b+16*g+r)+2] = clrmap16[b];
					kLogPal[4+4*(256*b+16*g+r)+3] = PC_NOCOLLAPSE;
				}
			}
		}
		kLogPal[2] = 0;
		kLogPal[3] = 16;
	} else {
		memcpy(kLogPal, OurPal, sizeof(OurPal));
	}
	kPal = CreatePalette((LPLOGPALETTE)kLogPal);
}

void DestroyPalette() {

	DeleteObject(kPal);
}
